# Standard library imports...
import logging
from requests import exceptions
import re
from urllib.parse import urlparse

# Third-party imports...
from flask import (
    Flask,
    abort,
    jsonify,
    request,
    make_response,
    render_template
)
from flask_httpauth import HTTPTokenAuth
from wtforms import (
    Form,
    StringField,
    PasswordField,
    validators
)

# Local imports...
from .matrix_api import create_account
from . import config
from . import tokens


app = Flask(__name__)

auth = HTTPTokenAuth(scheme='SharedSecret')
logger = logging.getLogger(__name__)


re_mxid = re.compile(r'^@?[a-zA-Z_\-=\.\/0-9]+(:[a-zA-Z\-\.:\/0-9]+)?$')


def validate_token(form, token):
    """
    validates token

    Parameters
    ----------
    arg1 : Form object
    arg2 : str
        token name, e.g. 'DoubleWizardSki'

    Raises
    -------
    ValidationError
        Token is invalid

    """
    tokens.tokens.load()
    if not tokens.tokens.valid(token.data):
        raise validators.ValidationError('Token is invalid')


def validate_username(form, username):
    """
    validates username

    Parameters
    ----------
    arg1 : Form object
    arg2 : str
        username name, e.g: '@user:matrix.org' or 'user'
        https://github.com/matrix-org/matrix-doc/blob/master/specification/appendices/identifier_grammar.rst#user-identifiers
    Raises
    -------
    ValidationError
        Username doesn't follow mxid requirements
    """
    domain = urlparse(config.config.server_location).hostname
    re_mxid = r'^@?[a-zA-Z_\-=\.\/0-9]+(:' + \
              re.escape(domain) + \
              r')?$'
    err = "Username doesn't follow pattern: '%s'" % re_mxid
    if not re.search(re_mxid, username.data):
        raise validators.ValidationError(err)


def validate_password(form, password):
    """
    validates username

    Parameters
    ----------
    arg1 : Form object
    arg2 : str
        password
    Raises
    -------
    ValidationError
        Password doesn't follow length requirements
    """
    min_length = config.config.password['min_length']
    err = 'Password should be between %s and 255 chars long' % min_length
    if len(password.data) < min_length or len(password.data) > 255:
        raise validators.ValidationError(err)


class RegistrationForm(Form):
    """
    Registration Form

    validates user account registration requests
    """
    username = StringField('Username', [
        validators.Length(min=1, max=200),
        # validators.Regexp(re_mxid)
        validate_username
    ])
    password = PasswordField('New Password', [
        # validators.Length(min=8),
        validate_password,
        validators.DataRequired(),
        validators.EqualTo('confirm', message='Passwords must match')
    ])
    confirm = PasswordField('Repeat Password')
    token = StringField('Token', [
        validators.Regexp(r'^([A-Z][a-z]+)+$'),
        validate_token
    ])


@auth.verify_token
def verify_token(token):
    return token == config.config.shared_secret


@auth.error_handler
def unauthorized():
    resp = {
                'errcode': 'MR_BAD_SECRET',
                'error': 'wrong shared secret'
            }
    return make_response(jsonify(resp), 401)


@app.route('/register', methods=['GET', 'POST'])
def register():
    """
    main user account registration endpoint
    to register an account you need to send a
    application/x-www-form-urlencoded request with
      - username
      - password
      - confirm
      - token
    as described in the RegistrationForm
    """
    if request.method == 'POST':
        logger.debug('an account registration started...')
        form = RegistrationForm(request.form)
        logger.debug('validating request data...')
        if form.validate():
            logger.debug('request valid')
            # remove sigil and the domain from the username
            username = form.username.data.rsplit(':')[0].split('@')[-1]
            logger.debug('creating account %s...' % username)
            # send account creation request to the hs
            try:
                account_data = create_account(form.username.data,
                                              form.password.data,
                                              config.config.server_location,
                                              config.config.shared_secret)
            except exceptions.ConnectionError:
                logger.error('can not connect to %s' % config.config.server_location,
                             exc_info=True)
                abort(500)
            except exceptions.HTTPError as e:
                resp = e.response
                error = resp.json()
                status_code = resp.status_code
                if status_code == 404:
                    logger.error('no HS found at %s' %
                                 config.config.server_location)
                elif status_code == 403:
                    logger.error(
                        'wrong shared registration secret or not enabled')
                elif status_code == 400:
                    # most likely this should only be triggered if a userid
                    # is already in use
                    return make_response(jsonify(error), 400)
                else:
                    logger.error('failure communicating with HS',
                                 exc_info=True)
                abort(500)
            logger.debug('account creation succeded!')
            tokens.tokens.use(form.token.data)
            return jsonify(access_token=account_data['access_token'],
                           home_server=account_data['home_server'],
                           user_id=account_data['user_id'],
                           status='success',
                           status_code=200)
        else:
            logger.debug('account creation failed!')
            resp = {'errcode': 'MR_BAD_USER_REQUEST',
                    'error': form.errors}
            return make_response(jsonify(resp), 400)
            # for fieldName, errorMessages in form.errors.items():
            #     for err in errorMessages:
            #         # return error to user
    else:
        server_name = config.config.server_name
        pw_length = config.config.password['min_length']
        return render_template('register.html',
                               server_name=server_name,
                               pw_length=pw_length,
                               riot_instance=config.config.riot_instance)


@app.route('/token', methods=['GET', 'POST'])
@auth.login_required
def token():
    tokens.tokens.load()

    data = False
    one_time = False
    ex_date = None
    if request.method == 'GET':
        return jsonify(tokens.tokens.toList())
    elif request.method == 'POST':
        data = request.get_json()
        if data:
            if 'ex_date' in data:
                ex_date = data['ex_date']
            if 'one_time' in data:
                one_time = data['one_time']
        try:
            token = tokens.tokens.new(ex_date=ex_date,
                                      one_time=one_time)
        except ValueError:
            resp = {
                'errcode': 'MR_BAD_DATE_FORMAT',
                'error': "date wasn't DD.MM.YYYY format"
            }
            return make_response(jsonify(resp), 400)
        return jsonify(token.toDict())
    abort(400)


@app.route('/token/<token>', methods=['GET', 'PUT'])
@auth.login_required
def token_status(token):
    tokens.tokens.load()
    data = False
    if request.method == 'GET':
        if tokens.tokens.get_token(token):
            return jsonify(tokens.tokens.get_token(token).toDict())
        else:
            resp = {
                'errcode': 'MR_TOKEN_NOT_FOUND',
                'error': 'token does not exist'
            }
            return make_response(jsonify(resp), 404)
    elif request.method == 'PUT':
        data = request.get_json(force=True)
        if data:
            if not data['disable']:
                resp = {
                    'errcode': 'MR_BAD_USER_REQUEST',
                    'error': 'PUT only allows "disable": true'
                }
                return make_response(jsonify(resp), 400)
            else:
                if tokens.tokens.disable(token):
                    return jsonify(tokens.tokens.get_token(token).toDict())
            resp = {
                'errcode': 'MR_TOKEN_NOT_FOUND',
                'error': 'token does not exist or is already disabled'
            }
            return make_response(jsonify(resp), 404)
    abort(400)



# Standard library imports...
import os
import site
import sys

# Third-party imports...
from appdirs import user_config_dir

__location__ = os.path.realpath(os.path.join(os.getcwd(),
                                os.path.dirname(__file__)))
WORD_LIST_PATH = os.path.join(__location__, 'wordlist.txt')
CONFIG_PATH1 = os.path.join(__location__, '../')
CONFIG_PATH2 = os.path.join(sys.prefix, 'config/')
CONFIG_PATH3 = os.path.join(user_config_dir('matrix-registration') + '/')
CONFIG_PATH4 = os.path.join(site.USER_BASE, 'config/')




# standard library imports...
import argparse
import logging
import logging.config

# third party imports...
from flask_limiter import Limiter
from flask_limiter.util import get_ipaddr
from flask_cors import CORS

# local imports...
from . import config
from . import tokens
from .api import app


def run_api(args):
    Limiter(
        app,
        key_func=get_ipaddr,
        default_limits=config.config.rate_limit
    )
    if config.config.allow_cors:
        CORS(app)
    app.run(host=config.config.host, port=config.config.port)


def generate_token(args):
    token = tokens.tokens.new(ex_date=args.expire, one_time=args.one_time)
    print(token.name)


def status_token(args):
    if args.disable:
        print(tokens.tokens.disable(args.disable))
    if args.status:
        print(tokens.tokens.get_token(args.status))
    if args.list:
        # print(json.dumps(tokens.tokens.__dict__,
        #                  sort_keys=True,
        #                  indent=4,
        #                  default=str))
        print(tokens.tokens)


parser = argparse.ArgumentParser(
    description='a token based matrix registration app',
    prog='python -m matrix_registration')
parser.add_argument('--config-path', default='config.yaml'
        , help='specifies the config file to be used', metavar='PATH')

# subparser
subparsers = parser.add_subparsers(
                        help='sub-commands. for ex. \'gen -h\' ' +
                        'for additional help')

# api-parser
parser_a = subparsers.add_parser('api', help='start as api')
parser_a.set_defaults(func=run_api)

# generate-parser
parser_g = subparsers.add_parser('gen',
                                 help='generate new token. ' +
                                 '-o onetime, -e expire date')
parser_g.add_argument('-o', '--one-time', action='store_true',
                      help='make token one-time-useable')
parser_g.add_argument('-e', '--expire', type=str, default=None,
                      help='expire date: DD.MM.YYYY')
parser_g.set_defaults(func=generate_token)

# status-parser
parser_s = subparsers.add_parser('status',
                                 help='view status or disable ' +
                                 'token. -s status, -d disable, -l list')
parser_s.add_argument('-s', '--status', type=str, default=None,
                      help='token status')
parser_s.add_argument('-l', '--list', action='store_true',
                      help='list tokens')
parser_s.add_argument('-d', '--disable', type=str, default=None,
                      help='disable token')
parser_s.set_defaults(func=status_token)

args = parser.parse_args()

config.config = config.Config(args.config_path)
logging.config.dictConfig(config.config.logging)
tokens.tokens = tokens.Tokens()

logger = logging.getLogger(__name__)

logger.debug('called with args: %s' % args)
if 'func' in args:
    args.func(args)



# Standard library imports...
# from collections import namedtuple
import logging
import os
import sys

# Third-party imports...
import yaml

# Local imports...
from .constants import (
    CONFIG_PATH1,
    CONFIG_PATH2,
    CONFIG_PATH3,
    CONFIG_PATH4
)
CONFIG_PATHS = [
    CONFIG_PATH1,
    CONFIG_PATH2,
    CONFIG_PATH3,
    CONFIG_PATH4
]
CONFIG_SAMPLE_NAME = "config.sample.yaml"
CONFIG_NAME = 'config.yaml'
logger = logging.getLogger(__name__)


class Config:
    """
    Config

    loads a dict or a yaml file to be accessible by all files in the module
    """
    def __init__(self, data):
        self.data = data
        self.CONFIG_PATH = None
        self.location = None
        self.load()

    def load(self):
        """
        loads the dict/the yaml file and recursively sets dictionary to class properties
        """
        logger.debug('loading config...')
        dictionary = None
        config_default = True
        if type(self.data) is dict:
            logger.debug('from dict...')
            dictionary = self.data
            config_default = False 
        else:
            logger.debug('from file...')
            # check work dir and all other pip install locations for config
            if os.path.isfile(self.data):
                config_default = False
            else:
                # provided file not found checking typical installation dirs
                config_exists = False
                for path in CONFIG_PATHS:
                    if os.path.isfile(path + CONFIG_NAME):
                        self.CONFIG_PATH = path
                        config_exists = True
                        config_default = False
                if not config_exists:
                    # no config exists, use sample config instead
                    # check typical installation dirs for sample configs
                    for path in CONFIG_PATHS:
                        if os.path.isfile(path + CONFIG_SAMPLE_NAME):
                            self.CONFIG_PATH = path
                            config_exists = True
                    # check if still no config found
                    if not config_exists:
                        sys.exit('could not find any configuration file!')
                    self.data = os.path.join(self.CONFIG_PATH, CONFIG_SAMPLE_NAME)
                else:
                    self.data = os.path.join(self.CONFIG_PATH, CONFIG_NAME)
            try:
                with open(self.data, 'r') as stream:
                    dictionary = yaml.load(stream, Loader=yaml.SafeLoader)
            except IOError as e:
                sys.exit(e)
        if config_default:
            self.read_config(dictionary)

        logger.debug('setting config...')
        # recusively set dictionary to class properties
        for k, v in dictionary.items():
            setattr(self, k, v)
        logger.debug('config set!')
        # self.x = namedtuple('config',
        #                     dictionary.keys())(*dictionary.values())

    def update(self, data):
        """
        resets all options and loads the new config

        Parameters
        ----------
        arg1 : dict or path to config file
        """
        logger.debug('updating config...')
        self.data = data
        self.CONFIG_PATH = None
        self.location = None
        self.load()
        logger.debug('config updated!')

    def read_config(self, dictionary):
        """
        asks the user how to set the essential options

        Parameters
        ----------
        arg1 : dict
            with sample values
        """
        # important keys that need to be changed
        keys = ['server_location', 'server_name', 'shared_secret', 'port']
        for key in keys:
            temp = dictionary[key]
            dictionary[key] = input('enter {}, e.g. {}\n'.format(key, temp))
            if not dictionary[key].strip():
                dictionary[key] = temp
        # write to config file
        new_config_path = self.CONFIG_PATH + CONFIG_NAME
        relative_path = os.path.relpath(self.CONFIG_PATH + CONFIG_NAME)
        with open(new_config_path, 'w') as stream:
            yaml.dump(dictionary, stream, default_flow_style=False)
            print('config file written to "%s"' % relative_path)


config = None



# Standard library imports...
from datetime import datetime
import logging
import random
import sqlite3

# Third-party imports...
from dateutil import parser

# Local imports...
from . import config
from .constants import WORD_LIST_PATH

sqlite3.register_adapter(bool, int)
sqlite3.register_converter('BOOLEAN', lambda v: bool(int(v)))


logger = logging.getLogger(__name__)


def random_readable_string(length=3, wordlist=WORD_LIST_PATH):
    with open(wordlist) as f:
        lines = f.read().splitlines()
        string = ''
        for _ in range(length):
            string += random.choice(lines).title()
    return string


class Token(object):
    def __init__(self, name=False, ex_date=None, one_time=False, used=0):
        if not ex_date or ex_date == 'None':
            self.ex_date = None
        else:
            self.ex_date = parser.parse(ex_date)
        if name:
            self.name = name
        else:
            self.name = random_readable_string()
        self.one_time = one_time
        self.used = used

    def __repr__(self):
        return self.name

    def toDict(self):
        _token = {
            'name': self.name,
            'used': self.used,
            'ex_date': self.ex_date,
            'one_time': bool(self.one_time),
            'valid': self.valid()
        }
        return _token

    def valid(self):
        expired = False
        if self.ex_date:
            expired = self.ex_date < datetime.now()
        used = bool(self.one_time and self.used > 0)

        return (not expired) and (not used)

    def use(self):
        if self.valid():
            self.used += 1
            return True
        return False

    def disable(self):
        if self.valid():
            self.ex_date = datetime(1, 1, 1)
            return True
        return False


class Tokens():
    def __init__(self):
        logger.info('connecting to %s' % config.config.db)
        self.conn = sqlite3.connect(config.config.db, check_same_thread=False)
        self.c = self.conn.cursor()
        self.tokens = {}

        logger.debug('creating table')
        self.c.execute('''CREATE TABLE IF NOT EXISTS tokens
                          (name TEXT UNIQUE,
                          ex_date TEXT,
                          one_time BOOLEAN,
                          used INT)''')
        self.conn.commit()

        self.load()

    def __repr__(self):
        result = ''
        for tokens_key in self.tokens:
            result += '%s, ' % tokens_key
        return result[:-2]

    def toList(self):
        _tokens = []
        for tokens_key in self.tokens:
            _tokens.append(self.tokens[tokens_key].toDict())
        return _tokens

    def update(self, token):
        sql = 'UPDATE tokens SET used=?, ex_date=? WHERE name=?'
        self.c.execute(sql, (token.used, str(token.ex_date), token.name))
        self.conn.commit()

    def load(self):
        logger.debug('loading tokens from db...')
        self.tokens = {}
        # Get tokens
        self.c.execute('SELECT * FROM tokens')
        for token in self.c.fetchall():
            logger.debug(token)
            # token[0]=name
            self.tokens[token[0]] = Token(name=token[0],
                                          ex_date=str(token[1]),
                                          one_time=token[2],
                                          used=token[3])
        logger.debug('token loaded!')

    def get_token(self, token_name):
        logger.debug('getting token by name: %s' % token_name)
        try:
            token = self.tokens[token_name]
        except KeyError:
            return False
        return token

    def valid(self, token_name):
        logger.debug('checking if "%s" is valid' % token_name)
        token = self.get_token(token_name)
        # if token exists
        if token:
            return token.valid()
        return False

    def use(self, token_name):
        logger.debug('using token: %s' % token_name)
        token = self.get_token(token_name)
        if token:
            if token.use():
                self.update(token)
                return True
        return False

    def disable(self, token_name):
        logger.debug('disabling token: %s' % token_name)
        token = self.get_token(token_name)
        if token:
            if token.disable():
                self.update(token)
                return True
            self.update(token)
        return False

    def new(self, ex_date=None, one_time=False):
        logger.debug(('creating new token, with options: one_time: {},' +
                     'ex_dates: {}').format(one_time, ex_date))
        token = Token(ex_date=ex_date, one_time=one_time)
        sql = '''INSERT INTO tokens (name, ex_date, one_time, used)
                     VALUES (?, ?, ?, ?)'''

        self.c.execute(sql, (token.name,
                             str(token.ex_date),
                             token.one_time,
                             token.used))
        self.tokens[token.name] = token
        self.conn.commit()

        return token


tokens = None



# Standard library imports...
import hashlib
import hmac
import requests

import logging

logger = logging.getLogger(__name__)


def create_account(user, password, server_location, shared_secret,
                   admin=False):
    """
    creates account
    https://github.com/matrix-org/synapse/blob/master/synapse/_scripts/register_new_matrix_user.py

    Parameters
    ----------
    arg1 : str
        local part of the new user
    arg2 : str
        password
    arg3 : str
        url to homeserver
    arg4 : str
        Registration Shared Secret as set in the homeserver.yaml
    arg5 : bool
        register new user as an admin.
    Raises
    -------
    requests.exceptions.ConnectionError:
        can't connect to homeserver
    requests.exceptions.HTTPError:
        something with the communciation to the homeserver failed
    """
    nonce = _get_nonce(server_location)

    mac = hmac.new(
        key=str.encode(shared_secret),
        digestmod=hashlib.sha1,
    )

    mac.update(nonce.encode())
    mac.update(b'\x00')
    mac.update(user.encode())
    mac.update(b'\x00')
    mac.update(password.encode())
    mac.update(b'\x00')
    mac.update(b'admin' if admin else b'notadmin')

    mac = mac.hexdigest()

    data = {
        'nonce': nonce,
        'username': user,
        'password': password,
        'admin': admin,
        'mac': mac,
    }

    server_location = server_location.rstrip('/')

    r = requests.post('%s/_matrix/client/r0/admin/register' % (server_location),
                      json=data)
    r.raise_for_status()
    return r.json()


def _get_nonce(server_location):
    r = requests.get('%s/_matrix/client/r0/admin/register' % (server_location))
    r.raise_for_status()
    return r.json()['nonce']



#!/usr/bin/env python
import codecs
import os
import re
import setuptools

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name='matrix-registration',
    version=find_version("matrix_registration", "__init__.py"),
    description='token based matrix registration app',
    author='Jona Abdinghoff (ZerataX)',
    author_email='mail@zera.tax',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/zeratax/matrix-registration",
    packages=['matrix_registration'],
    package_data={'matrix_registration': ['*.txt',
                                          'templates/*.html',
                                          'static/css/*.css',
                                          'static/images/*.jpg',
                                          'static/images/*.png',
                                          'static/images/*.ico']},
    include_package_data=True,
    python_requires='~=3.6',

    install_requires=[
        "appdirs==1.4.3",
        "Flask>=1.0.2",
        "flask-cors==3.0.7",
        "flask-httpauth==3.2.4",
        "flask-limiter==1.0.1",
        "python-dateutil>=2.7.3",
        "PyYAML>=5.1",
        "requests>=2.21.0",
        "WTForms>=2.2.1"
    ],
    tests_require=[
        "parameterized==0.7.0",
        "flake8==3.7.7"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Communications :: Chat",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    entry_points="""
        [console_scripts]
        matrix_registration=matrix_registration.__main__:main
    """,
    test_suite="tests.test_registration",
    data_files=[
        ("config", ["config.sample.yaml"]),
    ]
)
